from django.utils import timezone

from ...models import EherkenningConfiguration
from ...saml2.eherkenning import generate_eherkenning_metadata
from ._base import SamlMetadataBaseCommand


class Command(SamlMetadataBaseCommand):
    help = "Create the eHerkenning metadata file"
    config_model = EherkenningConfiguration
    default_certificate_label = "eHerkenning/eIDAS"

    def add_arguments(self, parser):
        super().add_arguments(parser)

        config: EherkenningConfiguration = self._get_config()

        parser.add_argument(
            "--loa",
            help="Level of Assurance (LoA) to use for all the services.",
            default="urn:etoegang:core:assurance-class:loa3",
        )
        parser.add_argument(
            "--eh-attribute-consuming-service-index",
            help="Attribute consuming service index for the eHerkenning service, defaults to 9052",
            default="9052",
        )
        parser.add_argument(
            "--eidas-attribute-consuming-service-index",
            help="Attribute consuming service index for the eHerkenning service, defaults to 9053",
            default="9053",
        )
        parser.add_argument(
            "--oin",
            required=not config.oin,
            default=config.oin,
            help="The OIN of the company providing the service.",
        )
        parser.add_argument(
            "--no-eidas",
            action="store_true",
            help="If True, then the service catalogue will contain only the eHerkenning service. Defaults to False.",
            default=False,
        )

    def get_filename(self):
        date_string = timezone.now().date().isoformat()
        return f"eherkenning-metadata-{date_string}.xml"

    def generate_metadata(self, options):
        return generate_eherkenning_metadata()



import uuid

from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils.translation import gettext_lazy as _

from ..choices import AssuranceLevels, XMLContentTypes
from ..validators import oin_validator
from .base import BaseConfiguration


class EherkenningConfiguration(BaseConfiguration):
    loa = models.CharField(
        _("LoA"),
        blank=True,
        choices=AssuranceLevels.choices,
        default=AssuranceLevels.substantial,
        help_text=_("Level of Assurance (LoA) to use for all the services."),
        max_length=100,
    )
    eh_attribute_consuming_service_index = models.CharField(
        _("eHerkenning attribute consuming service index"),
        blank=True,
        default="9052",
        help_text=_("Attribute consuming service index for the eHerkenning service"),
        max_length=100,
    )
    eh_requested_attributes = models.JSONField(
        _("requested attributes"),
        default=list,
        help_text=_(
            "A list of additional requested attributes. A single requested attribute "
            "can be a string (the name of the attribute) or an object with keys 'name' "
            "and 'required', where 'name' is a string and 'required' a boolean'."
        ),
    )
    eh_service_uuid = models.UUIDField(
        _("eHerkenning service UUID"),
        default=uuid.uuid4,
        help_text=_(
            "UUID of the eHerkenning service. Once entered into catalogues, changing "
            "the value is a manual process."
        ),
    )
    eh_service_instance_uuid = models.UUIDField(
        _("eHerkenning service UUID"),
        default=uuid.uuid4,
        help_text=_(
            "UUID of the eHerkenning service instance. Once entered into catalogues, "
            "changing the value is a manual process."
        ),
    )
    eidas_attribute_consuming_service_index = models.CharField(
        _("eIDAS attribute consuming service index"),
        blank=True,
        default="9053",
        help_text=_("Attribute consuming service index for the eIDAS service"),
        max_length=100,
    )
    eidas_requested_attributes = models.JSONField(
        _("requested attributes"),
        default=list,
        help_text=_(
            "A list of additional requested attributes. A single requested attribute "
            "can be a string (the name of the attribute) or an object with keys 'name' "
            "and 'required', where 'name' is a string and 'required' a boolean'."
        ),
    )
    eidas_service_uuid = models.UUIDField(
        _("eIDAS service UUID"),
        default=uuid.uuid4,
        help_text=_(
            "UUID of the eIDAS service. Once entered into catalogues, changing "
            "the value is a manual process."
        ),
    )
    eidas_service_instance_uuid = models.UUIDField(
        _("eIDAS service UUID"),
        default=uuid.uuid4,
        help_text=_(
            "UUID of the eHerkenning service instance. Once entered into catalogues, "
            "changing the value is a manual process."
        ),
    )
    oin = models.CharField(
        _("OIN"),
        help_text=_("The OIN of the company providing the service."),
        max_length=100,
        validators=[oin_validator],
    )
    no_eidas = models.BooleanField(
        _("no eIDAS"),
        blank=True,
        default=False,
        help_text=_(
            "If True, then the service catalogue will contain only the eHerkenning service."
        ),
    )
    privacy_policy = models.URLField(
        _("privacy policy"),
        help_text=_(
            "The URL where the privacy policy from the organization providing the "
            "service can be found."
        ),
        max_length=255,
    )
    makelaar_id = models.CharField(
        _("broker ID"),
        help_text=_("OIN of the broker used to set up eHerkenning/eIDAS."),
        max_length=100,
        validators=[oin_validator],
    )
    artifact_resolve_content_type = models.CharField(
        _("resolve artifact binding content type"),
        choices=XMLContentTypes.choices,
        default=XMLContentTypes.soap_xml,
        max_length=100,
        help_text=_(
            "'application/soap+xml' is considered legacy and modern brokers typically "
            "expect 'text/xml'."
        ),
    )
    service_language = models.CharField(
        _("service language"),
        max_length=2,
        default="nl",
        help_text=_("Metadata for eHerkenning/eidas will contain this language key"),
    )

    class Meta:
        verbose_name = _("Eherkenning/eIDAS configuration")

    def as_dict(self) -> dict:
        """
        Emit the configuration as a dictionary compatible with the old settings format.
        """
        organization = None

        if (
            not self.certificate
            or not self.certificate.private_key
            or not self.certificate.public_certificate
        ):
            raise ImproperlyConfigured(
                "No (valid) certificate configured. The configuration needs a "
                "certificate with private key and public certificate."
            )

        if self.organization_url and self.organization_name:
            organization = {
                self.service_language: {
                    "name": self.organization_name,
                    "displayname": self.organization_name,
                    "url": self.organization_url,
                }
            }

        # at least the EH service
        services = [
            {
                "service_uuid": str(self.eh_service_uuid),
                "service_name": self.service_name,
                "attribute_consuming_service_index": self.eh_attribute_consuming_service_index,
                "service_instance_uuid": str(self.eh_service_instance_uuid),
                "service_description": self.service_description,
                "service_url": self.base_url,
                "privacy_policy_url": self.privacy_policy,
                "herkenningsmakelaars_id": self.makelaar_id,
                "requested_attributes": self.eh_requested_attributes,
                "entity_concerned_types_allowed": [
                    {
                        "set_number": "1",
                        "name": "urn:etoegang:1.9:EntityConcernedID:RSIN",
                    },
                    {
                        "set_number": "1",
                        "name": "urn:etoegang:1.9:EntityConcernedID:KvKnr",
                    },
                    {
                        "set_number": "2",
                        "name": "urn:etoegang:1.9:EntityConcernedID:KvKnr",
                    },
                ],
                "language": self.service_language,
            }
        ]

        # add eIDAS
        if not self.no_eidas:
            eidas_service = {
                "service_uuid": str(self.eidas_service_uuid),
                "service_name": f"{self.service_name} (eIDAS)",
                "attribute_consuming_service_index": self.eidas_attribute_consuming_service_index,
                "service_instance_uuid": str(self.eidas_service_instance_uuid),
                "service_description": self.service_description,
                "service_url": self.base_url,
                "privacy_policy_url": self.privacy_policy,
                "herkenningsmakelaars_id": self.makelaar_id,
                "requested_attributes": self.eidas_requested_attributes,
                "entity_concerned_types_allowed": [
                    {
                        "name": "urn:etoegang:1.9:EntityConcernedID:Pseudo",
                        "set_number": "1",
                    },
                ],
                "language": self.service_language,
                "classifiers": ["eIDAS-inbound"],
            }
            services.append(eidas_service)

        return {
            "base_url": self.base_url,
            "entity_id": self.entity_id,
            "metadata_file": self.idp_metadata_file,
            "key_file": self.certificate.private_key,
            "cert_file": self.certificate.public_certificate,
            "service_entity_id": self.idp_service_entity_id,
            "oin": self.oin,
            "services": services,
            "loa": self.loa,
            # optional in runtime code
            "want_assertions_encrypted": self.want_assertions_encrypted,
            "want_assertions_signed": self.want_assertions_signed,
            "key_passphrase": self.key_passphrase or None,
            "signature_algorithm": self.signature_algorithm,
            "digest_algorithm": self.digest_algorithm or None,
            "technical_contact_person_telephone": self.technical_contact_person_telephone
            or None,
            "technical_contact_person_email": self.technical_contact_person_email
            or None,
            "organization": organization,
            "organization_name": self.organization_name,
            "artifact_resolve_content_type": self.artifact_resolve_content_type,
        }



# Generated by Django 3.2.16 on 2022-11-02 10:46

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("digid_eherkenning", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="digidconfiguration",
            options={"verbose_name": "Digid configuration"},
        ),
        migrations.AlterModelOptions(
            name="eherkenningconfiguration",
            options={"verbose_name": "Eherkenning/eIDAS configuration"},
        ),
        migrations.AlterField(
            model_name="digidconfiguration",
            name="base_url",
            field=models.URLField(
                help_text="Base URL of the application, without trailing slash.",
                max_length=100,
                verbose_name="base URL",
            ),
        ),
        migrations.AlterField(
            model_name="eherkenningconfiguration",
            name="base_url",
            field=models.URLField(
                help_text="Base URL of the application, without trailing slash.",
                max_length=100,
                verbose_name="base URL",
            ),
        ),
    ]



from django.db import models
from django.utils.translation import gettext_lazy as _

from onelogin.saml2.constants import OneLogin_Saml2_Constants


class SectorType(models.TextChoices):
    bsn = "s00000000", "BSN"
    sofi = "s00000001", "SOFI"


class DigestAlgorithms(models.TextChoices):
    sha1 = OneLogin_Saml2_Constants.SHA1, "SHA1"
    sha256 = OneLogin_Saml2_Constants.SHA256, "SHA256"
    sha384 = OneLogin_Saml2_Constants.SHA384, "SHA384"
    sha512 = OneLogin_Saml2_Constants.SHA512, "SHA512"


class SignatureAlgorithms(models.TextChoices):
    dsa_sha1 = OneLogin_Saml2_Constants.DSA_SHA1, "DSA_SHA1"
    rsa_sha1 = OneLogin_Saml2_Constants.RSA_SHA1, "RSA_SHA1"
    rsa_sha256 = OneLogin_Saml2_Constants.RSA_SHA256, "RSA_SHA256"
    rsa_sha384 = OneLogin_Saml2_Constants.RSA_SHA384, "RSA_SHA384"
    rsa_sha512 = OneLogin_Saml2_Constants.RSA_SHA512, "RSA_SHA512"


# ref: https://afsprakenstelsel.etoegang.nl/display/as/Level+of+assurance
class AssuranceLevels(models.TextChoices):
    non_existent = "urn:etoegang:core:assurance-class:loa1", _("Non existent (1)")
    low = "urn:etoegang:core:assurance-class:loa2", _("Low (2)")
    low_plus = "urn:etoegang:core:assurance-class:loa2plus", _("Low (2+)")
    substantial = "urn:etoegang:core:assurance-class:loa3", _("Substantial (3)")
    high = "urn:etoegang:core:assurance-class:loa4", _("High (4)")


class XMLContentTypes(models.TextChoices):
    soap_xml = OneLogin_Saml2_Constants.SOAP_XML, "application/soap+xml"
    text_xml = OneLogin_Saml2_Constants.TEXT_XML, "text/xml"



from django.urls import path

from . import views

app_name = "eherkenning"

urlpatterns = [
    path("login/", views.eHerkenningLoginView.as_view(), name="login"),
    path("acs/", views.eHerkenningAssertionConsumerServiceView.as_view(), name="acs"),
]



from django.urls import reverse

from furl import furl

from ..models import DigidConfiguration
from .base import BaseSaml2Client


def generate_digid_metadata() -> bytes:
    client = DigiDClient()
    client.saml2_setting_kwargs = {"sp_validation_only": True}
    return client.create_metadata()


class DigiDClient(BaseSaml2Client):
    cache_key_prefix = "digid"
    cache_timeout = 60 * 60  # 1 hour

    @property
    def conf(self) -> dict:
        if not hasattr(self, "_conf"):
            db_config = DigidConfiguration.get_solo()
            self._conf = db_config.as_dict()
            self._conf.setdefault("acs_path", reverse("digid:acs"))
        return self._conf

    def create_config_dict(self, conf):
        config_dict = super().create_config_dict(conf)
        if conf["slo"]:
            slo_url = furl(conf["base_url"]) / reverse("digid:slo-soap")
            response_url = furl(conf["base_url"]) / reverse("digid:slo-redirect")
            config_dict["sp"]["singleLogoutService"] = {
                # URL where the <LogoutRequest> from the IdP will be sent (IdP-initiated logout)
                "url": slo_url.url,
                "binding": "urn:oasis:names:tc:SAML:2.0:bindings:SOAP",
                # URL Location where the <LogoutResponse> from the IdP will be sent
                # (SP-initiated logout, reply)
                "responseUrl": response_url.url,
                "responseBinding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect",
            }
        return config_dict

    def create_config(self, config_dict):
        config_dict["security"].update(
            {
                # None sent for digi-id.
                "wantAttributeStatement": False,
                # For DigiD, if the Metadata file expires, we sent them an update. So
                # there is no need for an expiry date.
                "metadataValidUntil": "",
                "metadataCacheDuration": "",
                "requestedAuthnContextComparison": "minimum",
                "requestedAuthnContext": [
                    "urn:oasis:names:tc:SAML:2.0:ac:classes:MobileTwoFactorContract",
                ],
            }
        )
        return super().create_config(config_dict)

    def create_authn_request(self, request, return_to=None):
        return super().create_authn_request(
            request,
            return_to=return_to,
            is_passive=False,
            set_nameid_policy=False,
            name_id_value_req=None,
        )



from typing import Optional

from django.conf import settings
from django.contrib import auth, messages
from django.http import HttpResponseRedirect
from django.shortcuts import resolve_url
from django.utils.translation import gettext as _
from django.views.generic.base import TemplateView, View

from ..exceptions import SAML2Error, eHerkenningNoRSINError
from ..forms import SAML2Form
from ..saml2.eherkenning import (
    eHerkenningClient,
    generate_dienst_catalogus_metadata,
    generate_eherkenning_metadata,
)
from .base import get_redirect_url


class eHerkenningLoginView(TemplateView):
    template_name = "digid_eherkenning/post_binding.html"

    def get_relay_state(self):
        """
        TODO: It might be a good idea to sign the relay state.
        But I can't think of a way this could be abused, since
        we re-check the url when processed by the ACS.
        """
        redirect_to = self.request.GET.get("next", "")
        return get_redirect_url(self.request, redirect_to)

    def get_attribute_consuming_service_index(self) -> Optional[str]:
        attribute_consuming_service_index = self.request.GET.get(
            "attr_consuming_service_index"
        )
        return attribute_consuming_service_index

    #
    # TODO: It might be a good idea to change this to a post-verb.
    # I can't think of any realy attack-vectors, but seems like a good
    # idea anyways.
    #
    def get_context_data(self, **kwargs):
        context_data = super().get_context_data(**kwargs)
        client = eHerkenningClient()
        location, parameters = client.create_authn_request(
            self.request,
            attr_consuming_service_index=self.get_attribute_consuming_service_index(),
        )

        context_data.update(
            {
                "url": location,
                "form": SAML2Form(
                    initial={
                        "SAMLRequest": parameters["SAMLRequest"],
                        "RelayState": self.get_relay_state(),
                    }
                ),
            }
        )
        return context_data


class eHerkenningAssertionConsumerServiceView(View):
    login_url = None

    def get_login_url(self):
        url = self.get_redirect_url()
        if url:
            return url

        return resolve_url(settings.LOGIN_URL)

    def get_success_url(self):
        url = self.get_redirect_url()
        return url or resolve_url(settings.LOGIN_REDIRECT_URL)

    def get_redirect_url(self):
        redirect_to = self.request.GET.get("RelayState")
        return get_redirect_url(self.request, redirect_to)

    def handle_error(self, request, message):
        messages.error(request, message)
        login_url = self.get_login_url()
        return HttpResponseRedirect(login_url)

    def get(self, request):
        try:
            user = auth.authenticate(
                request=request, eherkenning=True, saml_art=request.GET.get("SAMLart")
            )
        except eHerkenningNoRSINError:
            return self.handle_error(
                request,
                _(
                    "No RSIN returned by eHerkenning. Login to eHerkenning did not succeed."
                ),
            )
        except SAML2Error:
            return self.handle_error(
                request, _("Login to eHerkenning did not succeed. Please try again.")
            )

        if user is None:
            return self.handle_error(
                request, _("Login to eHerkenning did not succeed. Please try again.")
            )

        auth.login(request, user)

        return HttpResponseRedirect(self.get_success_url())



from django.urls import path

from .models import DigidConfiguration, EherkenningConfiguration
from .saml2.digid import generate_digid_metadata
from .saml2.eherkenning import (
    generate_dienst_catalogus_metadata,
    generate_eherkenning_metadata,
)
from .views.metadata import MetadataView

app_name = "metadata"

urlpatterns = [
    path(
        "digid",
        MetadataView.as_view(
            config_model=DigidConfiguration,
            metadata_generator=generate_digid_metadata,
        ),
        name="digid",
    ),
    path(
        "eherkenning",
        MetadataView.as_view(
            config_model=EherkenningConfiguration,
            metadata_generator=generate_eherkenning_metadata,
        ),
        name="eherkenning",
    ),
    path(
        "eherkenning/dienstcatalogus",
        MetadataView.as_view(
            config_model=EherkenningConfiguration,
            metadata_generator=generate_dienst_catalogus_metadata,
        ),
        name="eh-dienstcatalogus",
    ),
]



from django.urls import path

from digid_eherkenning.mock.views.digid import (
    DigiDAssertionConsumerServiceMockView,
    DigiDLoginMockView,
    DigiDLogoutMockView,
)

"""
this is a mock replacement for the regular digid_urls.py
"""

app_name = "digid"

urlpatterns = (
    path("login/", DigiDLoginMockView.as_view(), name="login"),
    path("acs/", DigiDAssertionConsumerServiceMockView.as_view(), name="acs"),
    path("logout/", DigiDLogoutMockView.as_view(), name="logout"),
)



from django.db.models import Manager


class BaseDigidManager(Manager):
    def get_by_bsn(self, bsn):
        raise NotImplementedError

    def digid_create(self, bsn, **kwargs):
        raise NotImplementedError


class BaseeHerkenningManager(Manager):
    def get_by_rsin(self, rsin):
        raise NotImplementedError

    def eherkenning_create(self, rsin, **kwargs):
        raise NotImplementedError



import logging
import sys

from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend
from django.utils.translation import gettext_lazy as _

from onelogin.saml2.utils import OneLogin_Saml2_ValidationError

from .choices import SectorType
from .exceptions import eHerkenningNoRSINError
from .saml2.digid import DigiDClient
from .saml2.eherkenning import eHerkenningClient
from .utils import get_client_ip

logger = logging.getLogger(__name__)

UserModel = get_user_model()


class BaseBackend(ModelBackend):
    service_name = None
    error_messages = {
        "login_cancelled": _(
            "The %(service)s login from %(ip)s did not succeed or was cancelled."
        ),
        "login_error": _(
            "A technical error occurred from %(ip)s during %(service)s login."
        ),
        "login_success": _(
            "User %(user)s%(user_info)s from %(ip)s logged in using %(service)s"
        ),
    }

    def log_success(self, request, message):
        logger.info(message, extra={"request": request})

    def log_error(self, request, message, exception=None):
        """
        General technical errors, are logged using this method.
        """
        if exception is not None:
            logger.exception(message, extra={"request": request})
        else:
            logger.error(message, extra={"request": request})

    def log_auth_failed(self, request, message, exception=None):
        """
        Errors where the user cancelled the request, or where login
        failed are logged here
        """
        logger.info(message, extra={"request": request}, exc_info=exception)


class BaseSaml2Backend(BaseBackend):
    def handle_validation_error(self, request):
        e = sys.exc_info()[1]
        assert e is not None, "This method needs to be called from Exception context."

        if e.code == OneLogin_Saml2_ValidationError.STATUS_CODE_AUTHNFAILED:
            error_message = self.error_messages["login_cancelled"] % {
                "ip": get_client_ip(request),
                "service": self.service_name,
            }
            self.log_auth_failed(request, error_message, e)
        else:
            error_message = self.error_messages["login_error"] % {
                "ip": get_client_ip(request),
                "service": self.service_name,
            }
            self.log_error(request, error_message, e)


class BSNBackendMixin:
    def get_or_create_user_from_bsn(self, request, bsn):
        if bsn == "":
            self.log_error(request, self.error_messages["digid_no_bsn"])
            return

        created = False
        try:
            user = UserModel.digid_objects.get_by_bsn(bsn)
        except UserModel.DoesNotExist:
            user = UserModel.digid_objects.digid_create(bsn)
            created = True

        success_message = self.error_messages["login_success"] % {
            "user": str(user),
            "user_info": _(" (new account)") if created else "",
            "ip": get_client_ip(request),
            "service": self.service_name,
        }

        self.log_success(request, success_message)

        return user


class DigiDBackend(BSNBackendMixin, BaseSaml2Backend):
    service_name = "DigiD"
    error_messages = dict(
        BaseSaml2Backend.error_messages,
        **{
            "digid_no_bsn": _("Login failed due to no BSN being returned by DigiD."),
        }
    )

    def authenticate(self, request, digid=None, saml_art=None, errors=[]):
        # Note: the fucntion has side-effect: it modifies 'errors' parameter
        # It's a workaround to access auth errors outside the Backend
        errors.clear()

        if saml_art is None:
            return

        if not digid:
            return

        # Digid Stap 6 / 7 Artifact Resolution

        client = DigiDClient()

        try:
            response = client.artifact_resolve(request, saml_art)
        except OneLogin_Saml2_ValidationError as e:
            errors.append(e)
            self.handle_validation_error(request)

            return

        try:
            name_id = response.get_nameid()
        except OneLogin_Saml2_ValidationError as e:
            errors.append(e)
            self.handle_validation_error(request)

            return

        # TODO:
        # Make sure the IP-address we get back for the 'subject' matches the IP-address of the user.
        #
        # This is not a requirement, but is a good idea. See DigiD - 5.1 Controle op IP adressen
        #
        # if get_client_ip(request) != authn_statement.subject_locality.address:
        #     return

        sector_code, sectoral_number = name_id.split(":")

        # We only care about users with a BSN.
        if sector_code != SectorType.bsn:
            self.log_error(request, self.error_messages["digid_no_bsn"])
            return

        bsn = sectoral_number

        # Produce a user
        user = self.get_or_create_user_from_bsn(request, bsn)

        # DigiD requires a session of max 15 minutes. See DigiDCheck 2.2 T14 -- Sessieduur
        session_age = client.conf.get("session_age", None)
        if session_age is not None:
            request.session.set_expiry(session_age)

        return user


class eHerkenningBackend(BaseSaml2Backend):
    service_name = "eHerkenning"

    def get_legal_subject_id(self, attributes, name_qualifier):
        rsin = ""
        for attribute_value in attributes.get("urn:etoegang:core:LegalSubjectID", []):
            if not isinstance(attribute_value, dict):
                continue
            name_id = attribute_value["NameID"]
            if name_id and name_id["NameQualifier"] == name_qualifier:
                rsin = name_id["value"]
        return rsin

    def get_legal_subject_kvk(self, attributes):
        return self.get_legal_subject_id(
            attributes, "urn:etoegang:1.9:EntityConcernedID:KvKnr"
        )

    def get_legal_subject_rsin(self, attributes):
        return self.get_legal_subject_id(
            attributes, "urn:etoegang:1.9:EntityConcernedID:RSIN"
        )

    def get_company_name(self, attributes):
        company_names = attributes.get(
            "urn:etoegang:1.11:attribute-represented:CompanyName", []
        )

        return " ".join(company_names)

    def get_kvk_number(self, attributes):
        kvk_numbers = attributes.get(
            "urn:etoegang:1.11:attribute-represented:KvKnr", []
        )

        if len(kvk_numbers) > 1:
            logger.error("More than 1 KVK-number returned.")
        if len(kvk_numbers) == 0:
            logger.error("No KVK-number returned.")
            return ""

        return kvk_numbers[0]

    def get_or_create_user(self, request, saml_response, saml_attributes):
        rsin = self.get_legal_subject_rsin(saml_attributes)
        if rsin == "":
            error_message = "Login failed due to no RSIN being returned by eHerkenning."
            raise eHerkenningNoRSINError(error_message)

        created = False
        try:
            user = UserModel.eherkenning_objects.get_by_rsin(rsin)
        except UserModel.DoesNotExist:
            user = UserModel.eherkenning_objects.eherkenning_create(rsin)
            created = True

        success_message = self.error_messages["login_success"] % {
            "user": str(user),
            "user_info": " (new account)" if created else "",
            "ip": get_client_ip(request),
            "service": self.service_name,
        }

        self.log_success(request, success_message)

        return user, created

    def authenticate(self, request, eherkenning=None, saml_art=None):
        if saml_art is None:
            return

        if not eherkenning:
            return

        client = eHerkenningClient()
        try:
            response = client.artifact_resolve(request, saml_art)
        except OneLogin_Saml2_ValidationError:
            self.handle_validation_error(request)
            return

        try:
            attributes = response.get_attributes()
        except OneLogin_Saml2_ValidationError:
            self.handle_validation_error(request)
            return

        user, created = self.get_or_create_user(request, response, attributes)

        session_age = client.conf.get("session_age", None)
        if session_age is not None:
            request.session.set_expiry(session_age)

        return user



import logging
from typing import Callable, List

from django.conf import settings
from django.core.cache import cache
from django.utils import timezone

from furl import furl
from onelogin.saml2.auth import OneLogin_Saml2_Auth
from onelogin.saml2.constants import OneLogin_Saml2_Constants
from onelogin.saml2.errors import OneLogin_Saml2_Error, OneLogin_Saml2_ValidationError
from onelogin.saml2.idp_metadata_parser import OneLogin_Saml2_IdPMetadataParser
from onelogin.saml2.logout_response import OneLogin_Saml2_Logout_Response
from onelogin.saml2.settings import OneLogin_Saml2_Settings
from onelogin.saml2.soap_logout_request import Soap_Logout_Request
from onelogin.saml2.utils import OneLogin_Saml2_Utils
from onelogin.saml2.xml_utils import OneLogin_Saml2_XML

from digid_eherkenning.utils import get_client_ip

logger = logging.getLogger(__name__)


def create_saml2_request(base_url, request):
    #
    # Because there might be proxying done before finally
    # getting to the Django server and SERVER_NAME and SERVER_PORT in request.META
    # might not be set correctly, instead, we hard-code these parameters
    # based on settings.
    #
    # X-Forwarded-For is also not an option, because it only forwards the
    # IP-Address.
    #
    parsed_url = furl(base_url)
    return {
        "https": "on" if parsed_url.scheme == "https" else "off",
        "http_host": parsed_url.netloc,
        "script_name": request.META["PATH_INFO"],
        "get_data": request.GET.copy(),
        "post_data": request.POST.copy(),
        # Uncomment if using ADFS as IdP, https://github.com/onelogin/python-saml/pull/144
        # 'lowercase_urlencoding': True,
        "query_string": request.META["QUERY_STRING"],
        "body": request.body,
    }


def get_service_name(conf: dict) -> str:
    _service_name = conf.get("service_name", "")
    return _service_name["en"] if isinstance(_service_name, dict) else _service_name


def get_service_description(conf: dict) -> str:
    _service_description = conf.get("service_description", "")
    return (
        _service_description["en"]
        if isinstance(_service_description, dict)
        else _service_description
    )


def get_requested_attributes(conf: dict) -> List[dict]:
    # There needs to be a RequestedAttribute element where the name is the ServiceID
    # https://afsprakenstelsel.etoegang.nl/display/as/DV+metadata+for+HM
    requested_attributes = []
    for requested_attribute in conf.get("requested_attributes", []):
        if isinstance(requested_attribute, dict):
            requested_attributes.append(requested_attribute)
        else:
            requested_attributes.append(
                {
                    "name": requested_attribute,
                    "required": True,
                }
            )

    return requested_attributes


class BaseSaml2Client:
    cache_key_prefix = "saml2_"
    cache_timeout = 60 * 60  # 1 hour
    saml2_setting_kwargs: dict = {
        "custom_base_path": None,
    }

    def __init__(self, conf=None):
        self.authn_storage = AuthnRequestStorage(
            self.cache_key_prefix, self.cache_timeout
        )

    @property
    def conf(self):
        raise NotImplementedError("Subclasses must implement the 'conf' property")

    @property
    def saml2_settings(self):
        if not hasattr(self, "_saml2_settings"):
            self._saml2_settings = self.create_config(
                self.create_config_dict(conf=self.conf)
            )
        return self._saml2_settings

    def create_metadata(self) -> bytes:
        return self.saml2_settings.get_sp_metadata()

    def create_authn_request(
        self, request, return_to=None, attr_consuming_service_index=None, **kwargs
    ):
        saml2_request = create_saml2_request(self.conf["base_url"], request)
        saml2_auth = OneLogin_Saml2_Auth(
            saml2_request, old_settings=self.saml2_settings, custom_base_path=None
        )
        url, parameters = saml2_auth.login_post(
            return_to=return_to,
            attr_consuming_service_index=attr_consuming_service_index,
            **kwargs,
        )

        # Save the request ID so we can verify that we've sent
        # it when we receive the Artifact/ACS response.
        request_id = saml2_auth.get_last_request_id()
        self.authn_storage.store(request_id, get_client_ip(request))

        return url, parameters

    def artifact_resolve(self, request, saml_art):
        saml2_request = create_saml2_request(self.conf["base_url"], request)
        saml2_auth = OneLogin_Saml2_Auth(
            saml2_request, old_settings=self.saml2_settings, custom_base_path=None
        )
        response = saml2_auth.artifact_resolve(saml_art)

        self.verify_saml2_response(response, get_client_ip(request))

        return response

    def handle_post_response(self, request):
        saml2_request = create_saml2_request(self.conf["base_url"], request)

        saml2_auth = OneLogin_Saml2_Auth(
            saml2_request, old_settings=self.saml2_settings, custom_base_path=None
        )

        response = saml2_auth.post_response()

        self.verify_saml2_response(response, get_client_ip(request))

        return response

    def verify_saml2_response(self, response, client_ip_address):
        #
        # SAMLProf: 4.1.4.2 <Response> Usage
        #
        # If the containing message is in response to an <AuthnRequest>,
        # then the InResponseTo attribute MUST match the request's ID
        #
        in_response_to = response.get_in_response_to()
        authn_request = self.authn_storage.get(in_response_to)
        if authn_request is None:
            raise OneLogin_Saml2_ValidationError(
                f"The InResponseTo of the Response: {in_response_to}, is not a request id"
                "found in the request cache",
                OneLogin_Saml2_ValidationError.WRONG_INRESPONSETO,
            )

        #
        # This is not a mandatory check by the SAML specification. But seems
        # like a good idea to guard against various attacks.
        #
        authn_ip_address = authn_request["client_ip_address"]
        if authn_ip_address != client_ip_address:
            raise OneLogin_Saml2_ValidationError(
                f"A different IP address ({authn_ip_address})"
                f"was used when retrieving the AuthNRequest then for retrieving"
                f" the request to the ACS ({client_ip_address}).",
                OneLogin_Saml2_ValidationError.WRONG_INRESPONSETO,
            )

    def create_config(self, config_dict):
        """
        Convert to the format expected by the OneLogin SAML2 library.
        """
        return OneLogin_Saml2_Settings(config_dict, **self.saml2_setting_kwargs)

    def create_config_dict(self, conf):
        """
        Convert the settings specified in conf, which has the following format

        base_url: URL which is prefixed before any URL used by the SP we set up.
        entity_id: SAML2 Entity id of SP we set up.
        key_file: Django FieldFile with private key in PEM format
        cert_file: Django FieldFile with with certificate in PEM format.

        metadata_file: Django FieldFile with metadata file which specifies the IDP.
        service_entity_id: entity id used to find settings of IDP in metadata file.
        attribute_consuming_service_index
        service_name: Name of SP service we set up. This is used in metadata generation.
        requested_attributes: List of attributes which should be returned by the IDP.
        """
        service_name = get_service_name(conf)
        service_description = get_service_description(conf)
        requested_attributes = get_requested_attributes(conf)

        with conf["cert_file"].open("r") as cert_file, conf["key_file"].open(
            "r"
        ) as key_file:
            certificate = cert_file.read()
            privkey = key_file.read()

        assert not conf["base_url"].endswith(
            "/"
        ), "Base URL must not end with a trailing slash"
        acs_url = furl(conf["base_url"]) / conf["acs_path"]
        setting_dict = {
            "strict": True,
            "security": {
                "signMetadata": True,
                "authnRequestsSigned": True,
                "logoutRequestSigned": True,
                "logoutResponseSigned": True,
                "wantAssertionsEncrypted": conf.get("want_assertions_encrypted", False),
                "wantAssertionsSigned": conf.get("want_assertions_signed", False),
                "soapClientKey": conf["key_file"].path,
                "soapClientCert": conf["cert_file"].path,
                "soapClientPassphrase": conf.get("key_passphrase", None),
                # algorithm for requests with HTTP-redirect binding.
                # AuthnRequest with HTTP-POST uses RSA_SHA256, which is hardcoded in OneLogin_Saml2_Auth.login_post
                "signatureAlgorithm": conf.get(
                    "signature_algorithm", OneLogin_Saml2_Constants.RSA_SHA1
                ),
                "digestAlgorithm": conf.get("digest_algorithm"),
            },
            "debug": settings.DEBUG,
            # Service Provider Data that we are deploying.
            "sp": {
                # Identifier of the SP entity  (must be a URI)
                "entityId": conf["entity_id"],
                # Specifies info about where and how the <AuthnResponse> message MUST be
                # returned to the requester, in this case our SP.
                "assertionConsumerService": {
                    "url": acs_url.url,
                    "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Artifact",
                },
                # If you need to specify requested attributes, set a
                # attributeConsumingService per service. nameFormat, attributeValue and
                # friendlyName can be omitted
                "attributeConsumingService": {
                    "index": conf.get("attribute_consuming_service_index", "1"),
                    "serviceName": service_name,
                    "serviceDescription": service_description,
                    "requestedAttributes": [
                        {
                            "name": attr["name"],
                            "isRequired": True if attr["required"] else False,
                        }
                        for attr in requested_attributes
                    ],
                },
                "NameIDFormat": "urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified",
                "x509cert": certificate,
                "privateKey": privkey,
                "privateKeyPassphrase": conf.get("key_passphrase", None),
            },
        }

        # check if we need to add the idp
        metadata_file = conf["metadata_file"]
        if metadata_file:
            with metadata_file.open("r") as metadata_file:
                metadata = metadata_file.read()

            parsed_idp_metadata = OneLogin_Saml2_IdPMetadataParser.parse(
                metadata, entity_id=conf["service_entity_id"]
            )
            if "idp" not in parsed_idp_metadata:
                logger.warning(
                    "IDP with entity_id %s not found in metadata, excluding idp "
                    "from settings_dict",
                    conf["service_entity_id"],
                )
            else:
                setting_dict["idp"] = parsed_idp_metadata["idp"]

        telephone = conf.get("technical_contact_person_telephone")
        email = conf.get("technical_contact_person_email")
        if telephone and email:
            setting_dict["contactPerson"] = {
                "technical": {"telephoneNumber": telephone, "emailAddress": email}
            }

        organization = conf.get("organization")
        if organization:
            setting_dict["organization"] = organization

        return setting_dict

    def create_logout_request(self, request, return_to=None, name_id=None):
        """
        :returns: Redirection URL for HTTP-redirect binding
        """
        saml2_request = create_saml2_request(self.conf["base_url"], request)
        saml2_auth = OneLogin_Saml2_Auth(
            saml2_request, old_settings=self.saml2_settings
        )
        url = saml2_auth.logout(return_to=return_to, name_id=name_id)

        # store request_id for validation during SLO callback
        request_id = saml2_auth.get_last_request_id()
        request.session["logout_request_id"] = request_id

        return url

    def handle_logout_response(
        self, request, keep_local_session=False, delete_session_cb=None
    ) -> None:
        """
        process logout response from IdP with HTTP redirect binding
        """
        saml2_request = create_saml2_request(self.conf["base_url"], request)
        saml2_auth = OneLogin_Saml2_Auth(
            saml2_request, old_settings=self.saml2_settings
        )
        request_id = request.session.get("logout_request_id")
        saml2_auth.process_slo(
            request_id=request_id,
            keep_local_session=keep_local_session,
            delete_session_cb=delete_session_cb,
        )

        errors = saml2_auth.get_errors()
        if errors:
            raise OneLogin_Saml2_Error(
                ", ".join(errors),
                OneLogin_Saml2_Error.SAML_LOGOUTRESPONSE_INVALID,
            )

    def handle_logout_request(
        self,
        request,
        keep_local_session: bool = False,
        delete_session_cb: Callable = None,
    ) -> str:
        """
        process request from IdP to the logout callback endpoint with SOAP binding
        OneLogin_Saml2_Auth.process_slo can't be used here, because it doesn't support SOAP binding

        :param keep_local_session: When false will destroy the local session, otherwise
        will not destroy the local session

        :param delete_session_cb: Callback function which destroys local sessions
        :return: Logout response
        """
        saml2_request = create_saml2_request(self.conf["base_url"], request)
        post_body = saml2_request.get("body")

        # validate request
        if not post_body or post_body.decode() == "{}":
            message = "SAML LogoutRequest body not found."
            logger.error("Logout request from Digid failed: %s", message)
            return OneLogin_Saml2_XML.generate_soap_fault_message(message)

        status = OneLogin_Saml2_Constants.STATUS_SUCCESS
        logout_request = Soap_Logout_Request(self.saml2_settings, post_body)
        try:
            logout_request.validate()
        except (OneLogin_Saml2_Error, OneLogin_Saml2_ValidationError) as exc:
            logger.error("Logout request from Digid failed: %s", exc)
            status = OneLogin_Saml2_Constants.STATUS_RESPONDER
        else:
            # delete local session
            if not keep_local_session:
                OneLogin_Saml2_Utils.delete_local_session(delete_session_cb)

        # construct response
        in_response_to = logout_request.id
        response_builder = OneLogin_Saml2_Logout_Response(self.saml2_settings)
        response_builder.build(in_response_to, status=status)
        logout_response = response_builder.get_xml()

        security = self.saml2_settings.get_security_data()

        # Algorithm hardcoded in the same way as in other backend communication: auth.artifact_resolve
        if security["logoutResponseSigned"]:
            logout_response = OneLogin_Saml2_Utils.add_sign(
                logout_response,
                self.saml2_settings.get_sp_key(),
                self.saml2_settings.get_sp_cert(),
                key_passphrase=self.saml2_settings.get_sp_key_passphrase(),
                sign_algorithm=OneLogin_Saml2_Constants.RSA_SHA256,
                digest_algorithm=OneLogin_Saml2_Constants.SHA256,
            )

        if isinstance(logout_response, bytes):
            logout_response = logout_response.decode()

        soap_logout_response = OneLogin_Saml2_XML.add_soap_envelope(logout_response)

        return soap_logout_response


class AuthnRequestStorage:
    def __init__(self, cache_key_prefix, cache_timeout):
        self.cache_key_prefix = cache_key_prefix
        self.cache_timeout = cache_timeout

    def get_cache_key(self, request_id):
        return f"{self.cache_key_prefix}_{request_id}"

    def store(self, request_id, client_ip_address):
        """
        Save the request id and the ip address of the client in the cache.
        We use this later to check if this match if the user retrieves
        the response.
        """

        cache_key = self.get_cache_key(request_id)
        cache_value = {
            "current_time": timezone.now(),
            "client_ip_address": client_ip_address,
        }
        cache.set(cache_key, cache_value, self.cache_timeout)

    def get(self, request_id):
        cache_key = self.get_cache_key(request_id)
        cached_value = cache.get(cache_key)
        return cached_value



from django import forms
from django.utils.translation import gettext_lazy as _


class PasswordLoginForm(forms.Form):
    auth_name = forms.CharField(
        max_length=255, required=True, label=_("DigiD gebruikersnaam")
    )
    auth_pass = forms.CharField(
        max_length=255, required=True, label=_("Wachtwoord"), widget=forms.PasswordInput
    )
    remember_login = forms.CharField(
        label=_("Onthoud mijn DigiD gebruikersnaam"), required=False
    )



import logging
from urllib.parse import urlencode

from django.conf import settings
from django.contrib import auth, messages
from django.contrib.auth.views import LogoutView
from django.http import HttpResponseBadRequest, HttpResponseRedirect
from django.shortcuts import resolve_url
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views import View
from django.views.generic import RedirectView

from digid_eherkenning.mock import conf
from digid_eherkenning.views.base import get_redirect_url

logger = logging.getLogger(__name__)


class DigiDLoginMockView(RedirectView):
    """
    mock replacement for digid:login

    we do a simple redirect instead of a form POST like post_binding.html does
    """

    def get_redirect_url(self, *args, **kwargs):
        url = self.request.build_absolute_uri(conf.get_idp_login_url())

        next_url = self.request.GET.get("next") or self.request.build_absolute_uri(
            conf.get_success_url()
        )
        next_url = get_redirect_url(
            self.request, next_url, require_https=self.request.is_secure()
        )
        if not next_url:
            logger.debug("bad 'next' redirect parameter")
            return HttpResponseBadRequest("bad 'next' redirect parameter")

        cancel_url = self.request.GET.get("cancel") or self.request.build_absolute_uri(
            conf.get_cancel_url()
        )
        cancel_url = get_redirect_url(
            self.request, cancel_url, require_https=self.request.is_secure()
        )
        if not cancel_url:
            logger.debug("bad 'cancel' redirect parameter")
            return HttpResponseBadRequest("bad 'cancel' redirect parameter")

        params = {
            "next": next_url,
            "cancel": cancel_url,
            "acs": self.request.build_absolute_uri(reverse("digid:acs")),
        }
        return f"{url}?{urlencode(params)}"


class DigiDAssertionConsumerServiceMockView(View):
    """
    mock replacement for digid:acs
    """

    def get_login_url(self):
        """
        where to go after unsuccessful login
        """
        url = self.request.build_absolute_uri(conf.get_cancel_url())
        url = get_redirect_url(
            self.request, url, require_https=self.request.is_secure()
        )
        if url:
            return url

        return resolve_url(settings.LOGIN_URL)

    def get_success_url(self):
        """
        where to go after successful login
        """
        url = self.request.GET.get("next") or self.request.build_absolute_uri(
            conf.get_success_url()
        )
        url = get_redirect_url(
            self.request, url, require_https=self.request.is_secure()
        )
        return url or resolve_url(settings.LOGIN_REDIRECT_URL)

    def get(self, request):
        user = auth.authenticate(request=request, bsn=request.GET.get("bsn"))
        if user is None:
            message = _(
                "An error occurred in the communication with DigiD. "
                "Please try again later. If this error persists, please "
                "check the website https://www.digid.nl for the latest information."
            )
            messages.error(request, message)
            login_url = self.get_login_url()
            return HttpResponseRedirect(login_url)

        auth.login(request, user)

        return HttpResponseRedirect(self.get_success_url())


class DigiDLogoutMockView(LogoutView):
    # in the future we can render a custom template to emulate Digid logout page
    pass



from io import StringIO

from django.core.management import CommandError, call_command
from django.test import TestCase

import pytest
from lxml import etree
from privates.test import temp_private_root

from digid_eherkenning.models import DigidConfiguration
from digid_eherkenning.saml2.digid import generate_digid_metadata

from .conftest import DIGID_TEST_CERTIFICATE_FILE, DIGID_TEST_KEY_FILE
from .mixins import DigidMetadataMixin

NAME_SPACES = {
    "md": "urn:oasis:names:tc:SAML:2.0:metadata",
    "ds": "http://www.w3.org/2000/09/xmldsig#",
}


@temp_private_root()
class DigidMetadataManagementCommandTests(TestCase):
    def test_generate_metadata_all_options_specified(self):
        stdout = StringIO()

        call_command(
            "generate_digid_metadata",
            "--no-save-config",
            "--slo",
            stdout=stdout,
            want_assertions_encrypted=True,
            want_assertions_signed=True,
            key_file=str(DIGID_TEST_KEY_FILE),
            cert_file=str(DIGID_TEST_CERTIFICATE_FILE),
            signature_algorithm="http://www.w3.org/2001/04/xmldsig-more#rsa-sha256",
            digest_algorithm="http://www.w3.org/2001/04/xmlenc#sha256",
            entity_id="http://test-entity.id",
            base_url="http://test-entity.id",
            attribute_consuming_service_index="9050",
            service_name="Test Service Name",
            service_description="Test Service Description",
            technical_contact_person_telephone="06123123123",
            technical_contact_person_email="test@test.nl",
            organization_name="Test organisation",
            organization_url="http://test-organisation.nl",
            test=True,
        )

        output = stdout.getvalue()
        entity_descriptor_node = etree.XML(output)

        self.assertEqual(
            "http://test-entity.id", entity_descriptor_node.attrib["entityID"]
        )

        sspo_descriptor_node = entity_descriptor_node.find(
            ".//md:SPSSODescriptor",
            namespaces=NAME_SPACES,
        )

        self.assertEqual("true", sspo_descriptor_node.attrib["AuthnRequestsSigned"])
        self.assertEqual("true", sspo_descriptor_node.attrib["WantAssertionsSigned"])

        certificate_node = entity_descriptor_node.find(
            ".//ds:X509Certificate",
            namespaces=NAME_SPACES,
        )
        self.assertIn(
            "MIIC0DCCAbigAwIBAgIUEjGmfCGa1cOiTi+UKtDQVtySOHUwDQYJKoZIhvcNAQEL",
            certificate_node.text,
        )

        signature_algorithm_node = entity_descriptor_node.find(
            ".//ds:SignatureMethod",
            namespaces=NAME_SPACES,
        )
        self.assertEqual(
            "http://www.w3.org/2001/04/xmldsig-more#rsa-sha256",
            signature_algorithm_node.attrib["Algorithm"],
        )

        digest_algorithm_node = entity_descriptor_node.find(
            ".//ds:DigestMethod",
            namespaces=NAME_SPACES,
        )
        self.assertEqual(
            "http://www.w3.org/2001/04/xmlenc#sha256",
            digest_algorithm_node.attrib["Algorithm"],
        )

        assertion_consuming_service_node = entity_descriptor_node.find(
            ".//md:AssertionConsumerService",
            namespaces=NAME_SPACES,
        )
        self.assertEqual(
            "http://test-entity.id/digid/acs/",
            assertion_consuming_service_node.attrib["Location"],
        )

        attribute_consuming_service_node = entity_descriptor_node.find(
            ".//md:AttributeConsumingService",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("9050", attribute_consuming_service_node.attrib["index"])

        service_name_node = entity_descriptor_node.find(
            ".//md:ServiceName",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("Test Service Name", service_name_node.text)

        service_description_node = entity_descriptor_node.find(
            ".//md:ServiceDescription",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("Test Service Description", service_description_node.text)

        organisation_name_node = entity_descriptor_node.find(
            ".//md:OrganizationName",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("Test organisation", organisation_name_node.text)

        organisation_display_node = entity_descriptor_node.find(
            ".//md:OrganizationDisplayName",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("Test organisation", organisation_display_node.text)

        organisation_url_node = entity_descriptor_node.find(
            ".//md:OrganizationURL",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("http://test-organisation.nl", organisation_url_node.text)

        contact_person_node = entity_descriptor_node.find(
            ".//md:ContactPerson",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("technical", contact_person_node.attrib["contactType"])

        contact_email_node = entity_descriptor_node.find(
            ".//md:EmailAddress",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("test@test.nl", contact_email_node.text)

        contact_telephone_node = entity_descriptor_node.find(
            ".//md:TelephoneNumber",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("06123123123", contact_telephone_node.text)

        slo_nodes = entity_descriptor_node.findall(
            ".//md:SingleLogoutService",
            namespaces=NAME_SPACES,
        )
        self.assertEqual(len(slo_nodes), 2)
        slo_soap, slo_redirect = slo_nodes
        self.assertEqual(
            slo_soap.attrib["Location"], "http://test-entity.id/digid/slo/soap/"
        )
        self.assertEqual(
            slo_soap.attrib["Binding"], "urn:oasis:names:tc:SAML:2.0:bindings:SOAP"
        )
        self.assertEqual(
            slo_redirect.attrib["Location"], "http://test-entity.id/digid/slo/redirect/"
        )
        self.assertEqual(
            slo_redirect.attrib["Binding"],
            "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect",
        )

    def test_missing_required_properties(self):
        expected_error = (
            "Error: the following arguments are required: --key-file, --cert-file, "
            "--entity-id, --base-url, --service-name, --service-description, "
            "--save-config/--no-save-config"
        )

        with self.assertRaisesMessage(CommandError, expected_error):
            call_command("generate_digid_metadata")

    def test_contact_telephone_no_email(self):
        stdout = StringIO()

        call_command(
            "generate_digid_metadata",
            "--no-save-config",
            "--slo",
            want_assertions_encrypted=True,
            want_assertions_signed=True,
            key_file=str(DIGID_TEST_KEY_FILE),
            cert_file=str(DIGID_TEST_CERTIFICATE_FILE),
            entity_id="http://test-entity.id",
            base_url="http://test-entity.id",
            service_name="Test Service Name",
            service_description="Test Service Description",
            technical_contact_person_telephone="06123123123",
            test=True,
            stdout=stdout,
        )

        output = stdout.getvalue()
        entity_descriptor_node = etree.XML(output)

        contact_email_node = entity_descriptor_node.find(
            ".//md:EmailAddress",
            namespaces=NAME_SPACES,
        )
        contact_telephone_node = entity_descriptor_node.find(
            ".//md:TelephoneNumber",
            namespaces=NAME_SPACES,
        )

        self.assertIsNone(contact_email_node)
        self.assertIsNone(contact_telephone_node)

    def test_organisation_url_no_service(self):
        stdout = StringIO()

        call_command(
            "generate_digid_metadata",
            "--no-save-config",
            "--slo",
            stdout=stdout,
            want_assertions_encrypted=True,
            want_assertions_signed=True,
            key_file=str(DIGID_TEST_KEY_FILE),
            cert_file=str(DIGID_TEST_CERTIFICATE_FILE),
            entity_id="http://test-entity.id",
            base_url="http://test-entity.id",
            service_name="Test Service Name",
            service_description="Test Service Description",
            organization_url="http://test-organisation.nl",
            test=True,
        )

        output = stdout.getvalue()
        entity_descriptor_node = etree.XML(output)

        organisation_name_node = entity_descriptor_node.find(
            ".//md:OrganizationName",
            namespaces=NAME_SPACES,
        )
        organisation_display_node = entity_descriptor_node.find(
            ".//md:OrganizationDisplayName",
            namespaces=NAME_SPACES,
        )
        organisation_url_node = entity_descriptor_node.find(
            ".//md:OrganizationURL",
            namespaces=NAME_SPACES,
        )

        self.assertIsNone(organisation_name_node)
        self.assertIsNone(organisation_display_node)
        self.assertIsNone(organisation_url_node)

    def test_slo_not_supported(self):
        stdout = StringIO()

        call_command(
            "generate_digid_metadata",
            "--no-save-config",
            "--no-slo",
            stdout=stdout,
            want_assertions_encrypted=True,
            want_assertions_signed=True,
            key_file=str(DIGID_TEST_KEY_FILE),
            cert_file=str(DIGID_TEST_CERTIFICATE_FILE),
            entity_id="http://test-entity.id",
            base_url="http://test-entity.id",
            service_name="Test Service Name",
            service_description="Test Service Description",
            test=True,
        )

        output = stdout.getvalue()
        entity_descriptor_node = etree.XML(output)

        single_logout_service_node = entity_descriptor_node.find(
            ".//md:SingleLogoutService",
            namespaces=NAME_SPACES,
        )
        self.assertIsNone(single_logout_service_node)

    def test_management_command_and_update_config(self):
        stdout = StringIO()
        assert not DigidConfiguration.objects.exists()

        call_command(
            "generate_digid_metadata",
            "--save-config",
            "--want-assertions-encrypted",
            "--no-only-assertions-signed",
            ["--attribute-consuming-service-index", "1"],
            key_file=str(DIGID_TEST_KEY_FILE),
            cert_file=str(DIGID_TEST_CERTIFICATE_FILE),
            entity_id="http://test-entity.id",
            base_url="http://test-entity.id",
            service_name="Test Service Name",
            service_description="Test Service Description",
            stdout=stdout,
            test=True,
        )

        self.assertTrue(DigidConfiguration.objects.exists())
        config = DigidConfiguration.get_solo()
        self.assertTrue(config.want_assertions_encrypted)
        self.assertFalse(config.want_assertions_signed)
        self.assertEqual(config.service_name, "Test Service Name")
        self.assertEqual(config.service_description, "Test Service Description")
        self.assertEqual(config.attribute_consuming_service_index, "1")

        self.assertIsNotNone(config.certificate)

        with config.certificate.private_key.open("rb") as privkey:
            with DIGID_TEST_KEY_FILE.open("rb") as source_privkey:
                self.assertEqual(privkey.read(), source_privkey.read())

        with config.certificate.public_certificate.open("rb") as cert:
            with DIGID_TEST_CERTIFICATE_FILE.open("rb") as source_cert:
                self.assertEqual(cert.read(), source_cert.read())


@pytest.mark.django_db
def test_properties_in_db_config_not_required(digid_config):
    """
    Assert that required properties already configured don't cause problems.
    """
    digid_config.service_description = "CLI test"
    digid_config.save()
    try:
        call_command(
            "generate_digid_metadata",
            "--no-save-config",
            "--test",
            stdout=StringIO(),
        )
    except CommandError:
        pytest.fail("Database configuration is valid for management commands.")


@pytest.mark.usefixtures("digid_config", "temp_private_root")
class DigidMetadataGenerationTests(DigidMetadataMixin, TestCase):
    def test_generate_metadata_all_options_specified(self):
        self.digid_config.want_assertions_signed = True
        self.digid_config.signature_algorithm = (
            "http://www.w3.org/2001/04/xmldsig-more#rsa-sha256"
        )
        self.digid_config.digest_algorithm = "http://www.w3.org/2001/04/xmlenc#sha256"
        self.digid_config.attribute_consuming_service_index = "9050"
        self.digid_config.technical_contact_person_telephone = "06123123123"
        self.digid_config.technical_contact_person_email = "test@test.nl"
        self.digid_config.organization_name = "Test organisation"
        self.digid_config.organization_url = "http://test-organisation.nl"
        self.digid_config.save()

        digid_metadata = generate_digid_metadata()

        entity_descriptor_node = etree.XML(digid_metadata)

        self.assertEqual(
            "http://test-entity.id", entity_descriptor_node.attrib["entityID"]
        )

        sspo_descriptor_node = entity_descriptor_node.find(
            ".//md:SPSSODescriptor",
            namespaces=NAME_SPACES,
        )

        self.assertEqual("true", sspo_descriptor_node.attrib["AuthnRequestsSigned"])
        self.assertEqual("true", sspo_descriptor_node.attrib["WantAssertionsSigned"])

        certificate_node = entity_descriptor_node.find(
            ".//ds:X509Certificate",
            namespaces=NAME_SPACES,
        )
        self.assertIn(
            "MIIC0DCCAbigAwIBAgIUEjGmfCGa1cOiTi+UKtDQVtySOHUwDQYJKoZIhvcNAQEL",
            certificate_node.text,
        )

        signature_algorithm_node = entity_descriptor_node.find(
            ".//ds:SignatureMethod",
            namespaces=NAME_SPACES,
        )
        self.assertEqual(
            "http://www.w3.org/2001/04/xmldsig-more#rsa-sha256",
            signature_algorithm_node.attrib["Algorithm"],
        )

        digest_algorithm_node = entity_descriptor_node.find(
            ".//ds:DigestMethod",
            namespaces=NAME_SPACES,
        )
        self.assertEqual(
            "http://www.w3.org/2001/04/xmlenc#sha256",
            digest_algorithm_node.attrib["Algorithm"],
        )

        assertion_consuming_service_node = entity_descriptor_node.find(
            ".//md:AssertionConsumerService",
            namespaces=NAME_SPACES,
        )
        self.assertEqual(
            "http://test-entity.id/digid/acs/",
            assertion_consuming_service_node.attrib["Location"],
        )

        attribute_consuming_service_node = entity_descriptor_node.find(
            ".//md:AttributeConsumingService",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("9050", attribute_consuming_service_node.attrib["index"])

        service_name_node = entity_descriptor_node.find(
            ".//md:ServiceName",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("Test Service Name", service_name_node.text)

        service_description_node = entity_descriptor_node.find(
            ".//md:ServiceDescription",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("Test Service Description", service_description_node.text)

        organisation_name_node = entity_descriptor_node.find(
            ".//md:OrganizationName",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("Test organisation", organisation_name_node.text)

        organisation_display_node = entity_descriptor_node.find(
            ".//md:OrganizationDisplayName",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("Test organisation", organisation_display_node.text)

        organisation_url_node = entity_descriptor_node.find(
            ".//md:OrganizationURL",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("http://test-organisation.nl", organisation_url_node.text)

        contact_person_node = entity_descriptor_node.find(
            ".//md:ContactPerson",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("technical", contact_person_node.attrib["contactType"])

        contact_email_node = entity_descriptor_node.find(
            ".//md:EmailAddress",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("test@test.nl", contact_email_node.text)

        contact_telephone_node = entity_descriptor_node.find(
            ".//md:TelephoneNumber",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("06123123123", contact_telephone_node.text)

        slo_nodes = entity_descriptor_node.findall(
            ".//md:SingleLogoutService",
            namespaces=NAME_SPACES,
        )
        self.assertEqual(len(slo_nodes), 2)
        slo_soap, slo_redirect = slo_nodes
        self.assertEqual(
            slo_soap.attrib["Location"], "http://test-entity.id/digid/slo/soap/"
        )
        self.assertEqual(
            slo_soap.attrib["Binding"], "urn:oasis:names:tc:SAML:2.0:bindings:SOAP"
        )
        self.assertEqual(
            slo_redirect.attrib["Location"], "http://test-entity.id/digid/slo/redirect/"
        )
        self.assertEqual(
            slo_redirect.attrib["Binding"],
            "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect",
        )

    def test_contact_telephone_no_email(self):
        self.digid_config.technical_contact_person_telephone = "06123123123"
        self.digid_config.save()

        digid_metadata = generate_digid_metadata()
        entity_descriptor_node = etree.XML(digid_metadata)

        contact_email_node = entity_descriptor_node.find(
            ".//md:EmailAddress",
            namespaces=NAME_SPACES,
        )
        contact_telephone_node = entity_descriptor_node.find(
            ".//md:TelephoneNumber",
            namespaces=NAME_SPACES,
        )

        self.assertIsNone(contact_email_node)
        self.assertIsNone(contact_telephone_node)

    def test_organisation_url_no_service(self):
        self.digid_config.organization_url = "http://test-organisation.nl"
        self.digid_config.save()

        digid_metadata = generate_digid_metadata()
        entity_descriptor_node = etree.XML(digid_metadata)

        organisation_name_node = entity_descriptor_node.find(
            ".//md:OrganizationName",
            namespaces=NAME_SPACES,
        )
        organisation_display_node = entity_descriptor_node.find(
            ".//md:OrganizationDisplayName",
            namespaces=NAME_SPACES,
        )
        organisation_url_node = entity_descriptor_node.find(
            ".//md:OrganizationURL",
            namespaces=NAME_SPACES,
        )

        self.assertIsNone(organisation_name_node)
        self.assertIsNone(organisation_display_node)
        self.assertIsNone(organisation_url_node)

    def test_slo_not_supported(self):
        self.digid_config.slo = False
        self.digid_config.save()

        digid_metadata = generate_digid_metadata()

        entity_descriptor_node = etree.XML(digid_metadata)

        single_logout_service_node = entity_descriptor_node.find(
            ".//md:SingleLogoutService",
            namespaces=NAME_SPACES,
        )
        self.assertIsNone(single_logout_service_node)



from django.utils.http import url_has_allowed_host_and_scheme


def get_redirect_url(request, redirect_to, require_https=True):
    """
    Make sure the URL we redirect to is safe. HTTPs
    is always required.
    """

    url_is_safe = url_has_allowed_host_and_scheme(
        url=redirect_to,
        allowed_hosts=[
            request.get_host(),
        ],
        require_https=require_https,
    )
    if url_is_safe:
        return redirect_to

    return ""
